import base64
import gc
import sys
import pyarrow as pa
from uuid import UUID

from wonderwords import RandomWord


def arrow_iterator(table: pa.Table):
    # TODO: setting to 10k so it doesn't take too long?
    # Wrote this as a generator so we don't realize the whole table in memory
    reader = table.to_reader(max_chunksize=10_000)
    for batch in reader:
        for item in batch.to_pylist():
            yield item
    

def b64int(i: int):
    ibytes = int.to_bytes(i, length=8, byteorder='big').lstrip(b'\x00')
    if ibytes == b'':
        ibytes = b'\x00'
    return base64.b64encode(ibytes).decode('utf8').rstrip('=')


def get_random_name():
    random_words = RandomWord()
    return f"{random_words.word(include_parts_of_speech=['adjectives'])}-{random_words.word(include_parts_of_speech=['nouns'])}"


def assert_valid_project_id(project_id):
    try:
        uuid_obj = UUID(project_id, version=4)
    except ValueError:
        raise ValueError(f"`{project_id}` is not a valid project id.")


def get_object_size_in_bytes(obj):
    marked = {id(obj)}
    obj_q = [obj]
    sz = 0

    while obj_q:
        sz += sum(map(sys.getsizeof, obj_q))

        # Lookup all the object referred to by the object in obj_q.
        # See: https://docs.python.org/3.7/library/gc.html#gc.get_referents
        all_refr = ((id(o), o) for o in gc.get_referents(*obj_q))

        # Filter object that are already marked.
        # Using dict notation will prevent repeated objects.
        new_refr = {o_id: o for o_id, o in all_refr if o_id not in marked and not isinstance(o, type)}

        # The new obj_q will be the ones that were not marked,
        # and we will update marked with their ids so we will
        # not traverse them again.
        obj_q = new_refr.values()
        marked.update(new_refr.keys())

    return sz



"""
This class allows for programmatic interactions with Atlas - Nomic's neural database. Initialize AtlasClient in any Python context such as a script
or in a Jupyter Notebook to organize and interact with your unstructured data.
"""

import uuid
from typing import Dict, Iterable, List, Optional, Union

import pyarrow as pa
import numpy as np
from pandas import DataFrame
from loguru import logger
from tqdm import tqdm

from .project import AtlasProject
from .settings import *
from .utils import b64int, get_random_name, arrow_iterator


def map_embeddings(
    embeddings: np.array,
    data: List[Dict] = None,
    id_field: str = None,
    name: str = None,
    description: str = None,
    is_public: bool = True,
    colorable_fields: list = [],
    build_topic_model: bool = True,
    topic_label_field: str = None,
    num_workers: None = None,
    organization_name: str = None,
    reset_project_if_exists: bool = False,
    add_datums_if_exists: bool = False,
    shard_size: None = None,
    projection_n_neighbors: int = DEFAULT_PROJECTION_N_NEIGHBORS,
    projection_epochs: int = DEFAULT_PROJECTION_EPOCHS,
    projection_spread: float = DEFAULT_PROJECTION_SPREAD,
) -> AtlasProject:
    '''

    Args:
        embeddings: An [N,d] numpy array containing the batch of N embeddings to add.
        data: An [N,] element list of dictionaries containing metadata for each embedding.
        id_field: Specify your data unique id field. This field can be up 36 characters in length. If not specified, one will be created for you named `id_`.
        name: A name for your map.
        description: A description for your map.
        is_public: Should this embedding map be public? Private maps can only be accessed by members of your organization.
        colorable_fields: The project fields you want to be able to color by on the map. Must be a subset of the projects fields.
        organization_name: The name of the organization to create this project under. You must be a member of the organization with appropriate permissions. If not specified, defaults to your user accounts default organization.
        reset_project_if_exists: If the specified project exists in your organization, reset it by deleting all of its data. This means your uploaded data will not be contextualized with existing data.
        add_datums_if_exists: If specifying an existing project and you want to add data to it, set this to true.
        build_topic_model: Builds a hierarchical topic model over your data to discover patterns.
        topic_label_field: The metadata field to estimate topic labels from. Usually the field you embedded.
        projection_n_neighbors: The number of neighbors to build.
        projection_epochs: The number of epochs to build the map with.
        projection_spread: The spread of the map.

    Returns:
        An AtlasProject that now contains your map.

    '''

    assert isinstance(embeddings, np.ndarray), 'You must pass in a numpy array'

    if embeddings.size == 0:
        raise Exception("Your embeddings cannot be empty")

    if id_field is None:
        id_field = ATLAS_DEFAULT_ID_FIELD

    project_name = get_random_name()
    if description is None:
        description = 'A description for your map.'
    index_name = project_name

    if name:
        project_name = name
        index_name = name
    if description:
        description = description

    added_id_field = False
    if data is None:
        data = [{ATLAS_DEFAULT_ID_FIELD: b64int(i)} for i in range(len(embeddings))]
        added_id_field = True

    if id_field == ATLAS_DEFAULT_ID_FIELD and id_field not in data[0]:
        added_id_field = True
        for i in range(len(data)):
            # do not modify object the user passed in - also ensures IDs are unique if two input datums are the same *object*
            data[i] = data[i].copy()
            data[i][id_field] = b64int(i)

    if added_id_field:
        logger.warning("An ID field was not specified in your data so one was generated for you in insertion order.")

    project = AtlasProject(
        name=project_name,
        description=description,
        unique_id_field=id_field,
        modality='embedding',
        is_public=is_public,
        organization_name=organization_name,
        reset_project_if_exists=reset_project_if_exists,
        add_datums_if_exists=add_datums_if_exists,
    )

    number_of_datums_before_upload = project.total_datums

    # sends several requests to allow for threadpool refreshing. Threadpool hogs memory and new ones need to be created.
    logger.info("Uploading embeddings to Atlas.")

    if shard_size is not None:
        logger.warning("Passing `shard_size` is deprecated and will raise an error in a future release")
    if num_workers is not None:
        logger.warning("Passing `num_workers` is deprecated and will raise an error in a future release")

    try:
        project.add_embeddings(
            embeddings=embeddings,
            data=data,
        )
    except BaseException as e:
        if number_of_datums_before_upload == 0:
            logger.info(f"{project.name}: Deleting project due to failure in initial upload.")
            project.delete()
        raise e

    logger.info("Embedding upload succeeded.")

    # make a new index if there were no datums in the project before
    if number_of_datums_before_upload == 0:
        projection = project.create_index(
            name=index_name,
            colorable_fields=colorable_fields,
            build_topic_model=build_topic_model,
            projection_n_neighbors=projection_n_neighbors,
            projection_epochs=projection_epochs,
            projection_spread=projection_spread,
            topic_label_field=topic_label_field,
        )
        logger.info(str(projection))
    else:
        # otherwise refresh the maps
        project.rebuild_maps()

    project = project._latest_project_state()
    return project


def map_text(
    data: Union[Iterable[Dict], DataFrame],
    indexed_field: str,
    id_field: str = None,
    name: str = None,
    description: str = None,
    build_topic_model: bool = True,
    is_public: bool = True,
    colorable_fields: list = [],
    num_workers: None = None,
    organization_name: str = None,
    reset_project_if_exists: bool = False,
    add_datums_if_exists: bool = False,
    shard_size: None = None,
    projection_n_neighbors: int = DEFAULT_PROJECTION_N_NEIGHBORS,
    projection_epochs: int = DEFAULT_PROJECTION_EPOCHS,
    projection_spread: float = DEFAULT_PROJECTION_SPREAD,
    duplicate_detection: bool = False,
    duplicate_threshold: float = DEFAULT_DUPLICATE_THRESHOLD
) -> AtlasProject:
    '''
    Generates or updates a map of the given text.

    Args:
        data: An [N,] element iterable of dictionaries containing metadata for each embedding.
        indexed_field: The name the data field containing the text your want to map.
        id_field: Specify your data unique id field. This field can be up 36 characters in length. If not specified, one will be created for you named `id_`.
        name: A name for your map.
        description: A description for your map.
        build_topic_model: Builds a hierarchical topic model over your data to discover patterns.
        is_public: Should this embedding map be public? Private maps can only be accessed by members of your organization.
        colorable_fields: The project fields you want to be able to color by on the map. Must be a subset of the projects fields.
        organization_name: The name of the organization to create this project under. You must be a member of the organization with appropriate permissions. If not specified, defaults to your user account's default organization.
        reset_project_if_exists: If the specified project exists in your organization, reset it by deleting all of its data. This means your uploaded data will not be contextualized with existing data.
        add_datums_if_exists: If specifying an existing project and you want to add data to it, set this to true.
        projection_n_neighbors: The number of neighbors to build.
        projection_epochs: The number of epochs to build the map with.
        projection_spread: The spread of the map.

    Returns:
        The AtlasProject containing your map.

    '''
    if id_field is None:
        id_field = ATLAS_DEFAULT_ID_FIELD

    project_name = get_random_name()

    if description is None:
        description = 'A description for your map.'
    index_name = project_name

    if name:
        project_name = name
        index_name = name

    if isinstance(data, DataFrame):
        # Convert DataFrame to a generator of dictionaries
        data_iterator = (row._asdict() for row in data.itertuples(index=False))
    elif isinstance(data, pa.Table):
        # Create generator from pyarrow table
        data_iterator = arrow_iterator(data)
    else:
        data_iterator = iter(data)

    first_sample = next(data_iterator, None)

    if first_sample is None:
        logger.warning("Passed data has no samples. No project will be created")
        return

    project = AtlasProject(
        name=project_name,
        description=description,
        unique_id_field=id_field,
        modality='text',
        is_public=is_public,
        organization_name=organization_name,
        reset_project_if_exists=reset_project_if_exists,
        add_datums_if_exists=add_datums_if_exists,
    )

    add_id_field = False

    if id_field == ATLAS_DEFAULT_ID_FIELD and id_field not in first_sample:
        add_id_field = True

    if add_id_field:
        logger.warning("An ID field was not specified in your data so one was generated for you in insertion order.")

    project._validate_map_data_inputs(colorable_fields=colorable_fields, id_field=id_field, data_sample=first_sample)

    number_of_datums_before_upload = project.total_datums

    logger.info("Uploading text to Atlas.")
    if shard_size is not None:
        logger.warning("Passing 'shard_size' is deprecated and will be removed in a future release.")
    if num_workers is not None:
        logger.warning("Passing 'num_workers' is deprecated and will be removed in a future release.")
    try:
        upload_batch_size = 100_000
        id_to_add = 0
        if add_id_field:
            first_sample = first_sample.copy()
            first_sample[id_field] = b64int(id_to_add)
            id_to_add += 1
        
        batch = [first_sample]

        for d in data_iterator:
            if add_id_field:
                # do not modify object the user passed in - also ensures IDs are unique if two input datums are the same *object*
                d = d.copy()
                # necessary to persist change
                d[id_field] = b64int(id_to_add)
                id_to_add += 1
            batch.append(d)
            if len(batch) >= upload_batch_size:
                project.add_text(batch)
                batch = []

        if len(batch) > 0:
            project.add_text(batch)
        
    except BaseException as e:
        if number_of_datums_before_upload == 0:
            logger.info(f"{project.name}: Deleting project due to failure in initial upload.")
            project.delete()
        raise e

    logger.info("Text upload succeeded.")

    # make a new index if there were no datums in the project before
    if number_of_datums_before_upload == 0:
        projection = project.create_index(
            name=index_name,
            indexed_field=indexed_field,
            colorable_fields=colorable_fields,
            build_topic_model=build_topic_model,
            projection_n_neighbors=projection_n_neighbors,
            projection_epochs=projection_epochs,
            projection_spread=projection_spread,
            duplicate_detection=duplicate_detection,
            duplicate_threshold=duplicate_threshold,
        )
        logger.info(str(projection))
    else:
        # otherwise refresh the maps
        project.rebuild_maps()

    project = project._latest_project_state()
    return project



import json
import os
import time
from pathlib import Path

import click
import requests
from rich.console import Console

tenants = {
    'staging': {'frontend_domain': 'staging-atlas.nomic.ai', 'api_domain': 'staging-api-atlas.nomic.ai'},
    'production': {'frontend_domain': 'atlas.nomic.ai', 'api_domain': 'api-atlas.nomic.ai'},
}

nomic_base_path = Path.home() / '.nomic'


def validate_api_http_response(response):
    if response.status_code >= 500 and response.status_code < 600:
        raise Exception("Cannot contact establish a connection with Nomic services.")

    return response


def get_api_credentials(fn=None):
    if fn is None:
        fn = 'credentials'
    filepath = nomic_base_path / fn
    if not filepath.exists():
        raise ValueError("You have not configured your Nomic API token. Run `nomic login` to configure.")

    with open(filepath, 'r') as file:
        credentials = json.load(file)
        return credentials


def login(token, tenant='production'):
    environment = tenants[tenant]
    auth0_auth_endpoint = f"https://{environment['frontend_domain']}/cli-login"

    console = Console()
    style = "bold"
    if not token:
        console.print("Authenticate with the Nomic API", style=style, justify="center")
        console.print(auth0_auth_endpoint, style=style, justify="center")
        console.print(
            "Click the above link to retrieve your access token and then run `nomic login \[token]`",
            style=style,
            justify="center",
        )
        exit()

    # save credential
    if not nomic_base_path.exists():
        nomic_base_path.mkdir()

    response = requests.get('https://' + environment['api_domain'] + f"/v1/user/token/refresh/{token}")
    response = validate_api_http_response(response)

    if not response.status_code == 200:
        raise Exception("Could not authorize you with Nomic. Run `nomic login` to re-authenticate.")

    bearer_token = response.json()['access_token']
    with open(os.path.join(nomic_base_path, 'credentials'), 'w') as file:
        saved_credentials = {'refresh_token': token, 'token': bearer_token, 'tenant': tenant, 'expires': time.time() + 80000}
        if tenant == 'enterprise':
            saved_credentials |= environment
        json.dump(saved_credentials, file)


def refresh_bearer_token():
    credentials = get_api_credentials()
    if time.time() >= credentials['expires']:
        try:
            environment = tenants[credentials['tenant']]
        except KeyError:
            environment = credentials
        response = requests.get(
            'https://' + environment['api_domain'] + f"/v1/user/token/refresh/{credentials['refresh_token']}"
        )
        response = validate_api_http_response(response)

        if not response.status_code == 200:
            raise Exception("Could not authorize you with Nomic. Run `nomic login` to re-authenticate.")

        bearer_token = response.json()['access_token']
        credentials['token'] = bearer_token
        with open(os.path.join(nomic_base_path, 'credentials'), 'w') as file:
            json.dump(credentials, file)
    return credentials


def switch(tenant):
    assert tenant in ['staging', 'production', None]
    credentials = get_api_credentials()
    current_tenant = credentials['tenant']
    if tenant is None:
        print(f'Current tenant: {current_tenant}')
        return
    if current_tenant == tenant:
        return
    else:
        current_loc = nomic_base_path / 'credentials'
        new_loc = nomic_base_path / f'credentials_{current_tenant}'
        print(f'Switching from {current_tenant} to {tenant}.')
        if current_loc.exists():
            current_loc.rename(new_loc)
        if (nomic_base_path / f'credentials_{tenant}').exists():
            (nomic_base_path / f'credentials_{tenant}').rename(current_loc)
        else:
            login(token=None, tenant=tenant)


@click.command()
@click.option('--domain', default=None, help='Domain to use for Atlas enterprise login')
@click.argument('command', nargs=1, default='')
@click.argument('params', nargs=-1)
def cli(command, params, domain=None):
    if domain is not None:
        tenants['enterprise'] = { 'frontend_domain': domain, 'api_domain': f'api.{domain}' }
    if command == 'login':
        if len(params) == 0:
            login(token=None, tenant='production')
        if len(params) == 1 and params[0] == 'staging':
            login(token=None, tenant='staging')
        if len(params) == 2 and params[0] == 'staging':
            login(token=params[1], tenant='staging')
        if len(params) == 1 and params[0] == 'enterprise':
            if domain is None:
                raise ValueError('Must pass --domain to log into an enterprise environment')
            login(token=None, tenant='enterprise')
        if len(params) == 2 and params[0] == 'enterprise':
            if domain is None:
                raise ValueError('Must pass --domain to log into an enterprise environment')
            login(token=params[1], tenant='enterprise')
        if len(params) == 1:
            login(token=params[0], tenant='production')
    elif command == 'switch':
        if len(params) == 0:
            switch(tenant=None)
        if len(params) == 1:
            switch(tenant=params[0])
    else:
        raise ValueError(f"Command {command} not found.")


if __name__ == "__main__":
    cli()



import PIL.Image
import requests
import os
import base64
from io import BytesIO

from typing import List, Union
import PIL
from .project import AtlasClass


atlas_class = AtlasClass()



def text(texts: List[str], model: str = 'nomic-embed-text-v1'):
    """
    Generates embeddings for the given text.

    Args:
        texts: the text to embed
        model: the model to use when embedding

    Returns:
        An object containing your embeddings and request metadata
    """

    response = requests.post(
        atlas_class.atlas_api_path + "/v1/embedding/text",
        headers=atlas_class.header,
        json={'texts': texts, 'model': model},
    )

    if response.status_code == 200:
        return response.json()
    else:
        raise Exception(str(response.json()))

def images(images: Union[str, PIL.Image.Image], model: str = 'nomic-embed-vision-v1'):
    """
    Generates embeddings for the given images.

    Args:
        images: the images to embed. Can be file paths to images, image-file bytes or Pillow objects
        model: the model to use when embedding

    Returns:
        An object containing your embeddings and request metadata
    """

    batch_size = 250

    def run_inference(batch):
        response = requests.post(
            atlas_class.atlas_api_path + "/v1/embedding/image",
            headers=atlas_class.header,
            data={"model": "nomic-embed-vision-v1"},
            files=batch
        )

        if response.status_code == 200:
            return response.json()
        else:
            print(response.text)
            raise Exception(response.status_code)


    #naive batching, we should parallelize this across threads like we do with uploads.
    #TODO this should all be re-written prior to public release
    responses = []
    for i in range(0, len(images), batch_size):
        image_batch = []

        # process images into base64 encoded strings (for now)
        for image in images[i:i + batch_size]:
            #TODO implement check for bytes.
            #TODO implement check for a valid image.
            if isinstance(image, str) and os.path.exists(image):
                image_batch.append(('images', open(image, "rb")))

            elif isinstance(image, PIL.Image.Image):
                buffered = BytesIO()
                image.save(buffered, format="JPEG")
                image_batch.append(('images', buffered.getvalue()))
            else:
                raise ValueError(f"Not a valid file: {image}")
        print(len(image_batch))
        response = run_inference(batch=image_batch)
        print(response['usage'])
        responses.append(response)

    final_response = {}
    final_response['embeddings'] = [embedding for response in responses for embedding in response['embeddings']]
    final_response['usage'] = {}
    final_response['usage']['prompt_tokens'] = sum([response['usage']['prompt_tokens'] for response in responses])
    final_response['usage']['total_tokens'] = final_response['usage']['prompt_tokens']

    return final_response












from typing import Any, Dict, List, Union

import pyarrow as pa


def from_list(values: Dict[str, Any], schema=None) -> pa.Table:
    tb = pa.Table.from_pylist(values, schema=schema)
    return tb


permitted_types = {
    'integer': pa.int32(),
    'float': pa.float32(),
    'date': pa.timestamp('ms'),
    'string': pa.string(),
    'categorical': pa.string(),
}


def convert_pyarrow_schema_for_atlas(schema: pa.Schema) -> pa.Schema:
    """
    Convert a pyarrow schema to one with types that match the subset of types supported by Atlas for upload.
    """
    types = {}
    whitelist = {}
    for field in schema:
        if field.name.startswith('_'):
            # Underscore fields are private to Atlas and will be handled with their own logic.
            if not field.name in {"_embeddings"}:
                raise ValueError(f"Underscore fields are reserved for Atlas internal use: {field.name}")
            whitelist[field.name] = field.type
        elif pa.types.is_boolean(field.type):
            raise TypeError(f"Boolean type not supported: {field.name}")
        elif pa.types.is_list(field.type):
            raise TypeError(f"List types not supported: {field.name}")
        elif pa.types.is_struct(field.type):
            raise TypeError(f"Struct types not supported: {field.name}")
        elif pa.types.is_dictionary(field.type):
            types[field.name] = 'categorical'
        elif pa.types.is_string(field.type):
            types[field.name] = 'string'
        elif pa.types.is_integer(field.type):
            types[field.name] = 'integer'
        elif pa.types.is_floating(field.type):
            types[field.name] = 'float'
        elif pa.types.is_timestamp(field.type):
            types[field.name] = 'date'
        elif pa.types.is_temporal(field.type):
            types[field.name] = 'date'
        else:
            raise TypeError(f"Unknown type: {field.name} {field.type}")
    usertypes = {k: permitted_types[v] for k, v in types.items()}

    return pa.schema({**usertypes, **whitelist})
